@@ -3,7 +3,7 @@
 # Programmer friendly subprocess wrapper.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: May 4, 2018
+# Last Change: October 11, 2018
 # URL: https://executor.readthedocs.io
 
 """
@@ -53,7 +53,7 @@ import sys
 import tempfile
 
 # External dependencies.
-from humanfriendly import compact, concatenate, format
+from humanfriendly import compact, concatenate, format, pluralize
 from humanfriendly.terminal import connected_to_terminal
 from property_manager import (
     PropertyManager,
@@ -68,7 +68,7 @@ from six import string_types, text_type
 from executor.process import ControllableProcess
 
 # Semi-standard module versioning.
-__version__ = '19.3'
+__version__ = '21.2'
 
 # Initialize a logger.
 logger = logging.getLogger(__name__)
@@ -130,7 +130,7 @@ def execute(*command, **options):
              nonzero exit code (and :attr:`~ExternalCommand.check` is
              :data:`True`).
 
-    If :attr:`~ExternalCommand.async` is :data:`True` then :func:`execute()`
+    If :attr:`~ExternalCommand.asynchronous` is :data:`True` then :func:`execute()`
     will automatically start the external command for you using
     :func:`~ExternalCommand.start()` (but it won't wait for it to end). If you
     want to create an :class:`ExternalCommand` object instance without
@@ -183,7 +183,7 @@ def execute_prepared(command):
                     example :class:`.RemoteCommand`).
     :returns: The return value of this function depends on several options:
 
-              - If :attr:`~ExternalCommand.async` is :data:`True` the
+              - If :attr:`~ExternalCommand.asynchronous` is :data:`True` the
                 constructed :class:`ExternalCommand` object is returned.
 
               - If :attr:`~ExternalCommand.callback` is set the value of
@@ -196,7 +196,7 @@ def execute_prepared(command):
                 returned.
     :raises: See :func:`execute()` and :func:`.remote()`.
     """
-    if command.async:
+    if command.asynchronous:
         command.start()
         return command
     else:
@@ -241,6 +241,21 @@ class ExternalCommand(ControllableProcess):
         command. If an exception isn't already being raised
         :func:`check_errors()` is called to make sure the external command
         succeeded.
+
+    .. _event callbacks:
+
+    **Event callbacks**
+        The :attr:`start_event`, :attr:`retry_event` and :attr:`finish_event`
+        properties can be set to callbacks (callable values like functions) to
+        subscribe to the corresponding events. The callback receives a single
+        positional argument which is the :class:`ExternalCommand` object.
+
+        The :attr:`start_event` and :attr:`finish_event` properties were
+        originally created for use in command pools, for example to report to
+        the operator when specific commands are started and when they finish.
+        The event handling is performed inside the :class:`ExternalCommand`
+        class though, so you're free to repurpose these events outside the
+        context of command pools.
     """
 
     def __init__(self, *command, **options):
@@ -250,7 +265,7 @@ class ExternalCommand(ControllableProcess):
         :param command: Any positional arguments are converted to a list and
                         used to set :attr:`command`.
         :param options: Keyword arguments can be used to conveniently override
-                        the default values of :attr:`async`, :attr:`callback`,
+                        the default values of :attr:`asynchronous`, :attr:`callback`,
                         :attr:`capture`, :attr:`capture_stderr`, :attr:`check`,
                         :attr:`directory`, :attr:`encoding`,
                         :attr:`environment`, :attr:`fakeroot`, :attr:`input`,
@@ -276,7 +291,7 @@ class ExternalCommand(ControllableProcess):
         self.stderr_stream = CachedStream(self, 'stderr')
 
     @mutable_property
-    def async(self):
+    def asynchronous(self):
         """
         Enable asynchronous command execution.
 
@@ -304,6 +319,8 @@ class ExternalCommand(ControllableProcess):
           By using a temporary file the external command can produce output as
           fast or slow as it pleases without needing a thread or subprocess on
           our side to consume the output in real time.
+
+        .. seealso:: :attr:`async` (backwards compatible alias)
         """
         return False
 
@@ -315,19 +332,19 @@ class ExternalCommand(ControllableProcess):
         The value of :attr:`buffer_size` becomes the `bufsize` argument
         that's passed to :class:`subprocess.Popen` by :func:`start()`.
 
-        If :data:`async` is :data:`True` and :attr:`buffered` is :data:`False`
+        If :data:`asynchronous` is :data:`True` and :attr:`buffered` is :data:`False`
         the value of :attr:`buffer_size` defaults to 0 which means unbuffered,
         in all other cases its value defaults to -1 which means to use the
         system default buffer size.
         """
-        return 0 if self.async and not self.buffered else -1
+        return 0 if self.asynchronous and not self.buffered else -1
 
     @mutable_property
     def buffered(self):
         """
         Control whether command output is buffered to temporary files.
 
-        When :attr:`async` is :data:`True` and the standard output and/or error
+        When :attr:`asynchronous` is :data:`True` and the standard output and/or error
         streams are being captured, temporary files will be used to collect the
         output. This enables the use of the :attr:`output`, :attr:`stdout` and
         :attr:`stderr` properties to easily get the full output of the command
@@ -356,7 +373,7 @@ class ExternalCommand(ControllableProcess):
            known_states = set(['BLANK', 'LOCK', 'UNBLANK'])
 
            while True:
-               options = dict(async=True, capture=True, buffered=False)
+               options = dict(asynchronous=True, capture=True, buffered=False)
                with execute('xscreensaver-command', '-watch', **options) as command:
                    for line in command:
                        tokens = line.split()
@@ -429,8 +446,8 @@ class ExternalCommand(ControllableProcess):
 
         If this option is :data:`True` (the default) and the external command
         exits with a nonzero status code :exc:`ExternalCommandFailed` will be
-        raised by :func:`start()` (when :attr:`async` isn't set) or
-        :func:`wait()` (when :attr:`async` is set).
+        raised by :func:`start()` (when :attr:`asynchronous` isn't set) or
+        :func:`wait()` (when :attr:`asynchronous` is set).
         """
         return True
 
@@ -644,17 +661,7 @@ class ExternalCommand(ControllableProcess):
 
     @mutable_property
     def finish_event(self):
-        """
-        Optional callback that's called just after the command finishes.
-
-        The :attr:`start_event` and :attr:`finish_event` properties were
-        created for use in command pools, for example to report to the operator
-        when specific commands are started and when they finish. The invocation
-        of the :attr:`start_event` and :attr:`finish_event` callbacks is
-        performed inside the :class:`ExternalCommand` class though, so you're
-        free to repurpose these callbacks outside the context of command
-        pools.
-        """
+        """Optional callback that's called just after the command finishes (see `event callbacks`_)."""
 
     @mutable_property
     def group_by(self):
@@ -728,19 +735,30 @@ class ExternalCommand(ControllableProcess):
     @property
     def ionice_command(self):
         """The ionice_ command based on :attr:`ionice` (a list of strings)."""
-        return ['ionice', '--class', self.ionice] if self.ionice else []
+        return ['ionice', '-c', self.ionice] if self.ionice else []
 
     @property
     def is_finished(self):
         """
-        Whether the external command has finished execution.
+        Whether the external command has finished execution (excluding retries).
 
         :data:`True` once the external command has been started and has since
-        finished, :data:`False` when the external command hasn't been started
-        yet or is still running.
+        finished (excluding retries), :data:`False` when the external command
+        hasn't been started yet or is still running.
         """
         return self.was_started and not self.is_running
 
+    @property
+    def is_finished_with_retries(self):
+        """
+        Whether the external command has finished execution (including retries).
+
+        :data:`True` once the external command has been started and has since
+        finished (including retries), :data:`False` when the external command
+        hasn't been started yet, is still running or can be retried.
+        """
+        return self.is_finished and not self.retry_allowed
+
     @property
     def is_running(self):
         """:data:`True` if the process is currently running, :data:`False` otherwise."""
@@ -816,14 +834,15 @@ class ExternalCommand(ControllableProcess):
 
         When the :attr:`silent` option was originally added to executor it was
         implemented by redirecting the output streams to :data:`os.devnull`,
-        similar to how ``command &> /dev/null` works in Bash.
+        similar to how ``command &> /dev/null`` works in Bash.
 
         Since I made that decision I've regretted it many times because I ran
         into situations where :attr:`check` and :attr:`silent` were both set
         and :exc:`ExternalCommandFailed` was raised but I had no way to
         determine what had gone wrong.
 
-        This is why the :attr:`really_silent` property was introduced:
+        This is why the :attr:`really_silent` property was introduced in
+        executor release 19.0:
 
         - When :attr:`silent` is :data:`True` and :attr:`check` is
           :data:`False` the value of :attr:`really_silent` will be
@@ -861,6 +880,73 @@ class ExternalCommand(ControllableProcess):
                 self.wait()
             return self.callback(self)
 
+    @mutable_property
+    def retry(self):
+        """
+        Whether the external command should be retried when it fails (a boolean, defaults to :data:`False`).
+
+        .. warning:: Retrying of failing commands is an experimental feature
+                     that was introduced with the release of executor 20.0.
+                     Please refer to the `20.0 release notes`_ for details.
+
+        .. _20.0 release notes: https://executor.readthedocs.io/en/latest/changelog.html#release-20-0-2018-05-21
+        """
+        return False
+
+    @mutable_property
+    def retry_allowed(self):
+        """
+        :data:`True` if the external command can be retried, :data:`False` otherwise.
+
+        The value of this property is computed by checking if the following
+        conditions hold:
+
+        - :attr:`retry` is :data:`True`,
+        - :attr:`failed` is :data:`True`,
+        - :attr:`returncode` is not :data:`COMMAND_NOT_FOUND_STATUS`,
+        - :attr:`retry_count` is lower than :attr:`retry_limit`
+          (only if :attr:`retry_limit` is not zero).
+
+        Note that when the :attr:`retry_event` callback returns :data:`False`
+        to cancel the retrying of a failed command, the computed value of
+        :attr:`retry_allowed` is overridden by assigning :attr:`retry_allowed`
+        the value :data:`False`.
+        """
+        return (self.retry and self.failed and
+                self.returncode != COMMAND_NOT_FOUND_STATUS and
+                (self.retry_limit == 0 or self.retry_count < self.retry_limit))
+
+    @mutable_property
+    def retry_count(self):
+        """
+        The number of times that the command was retried (an integer number, defaults to 0).
+
+        The value of :attr:`retry_count` is automatically incremented by
+        :func:`start_once()` when it notices that :attr:`was_started` is
+        :data:`True` before :func:`start_once()` has started the command.
+        """
+        return 0
+
+    @mutable_property
+    def retry_event(self):
+        """
+        Optional callback that's called when a command is retried  (see `event callbacks`_).
+
+        The callback can return :data:`False` to abort retrying.
+        """
+
+    @mutable_property
+    def retry_limit(self):
+        """
+        The maximum number of times to *retry* the command when it fails (an integer, defaults to 2).
+
+        Given the default value of two, when :attr:`retry` is :data:`True`
+        the command will be run at most three times (the initial run and
+        two retries). The value 0 means the command will be retried until
+        it succeeds.
+        """
+        return 2
+
     @mutable_property
     def returncode(self):
         """
@@ -907,17 +993,7 @@ class ExternalCommand(ControllableProcess):
 
     @mutable_property
     def start_event(self):
-        """
-        Optional callback that's called just before the command is started.
-
-        The :attr:`start_event` and :attr:`finish_event` properties were
-        created for use in command pools, for example to report to the operator
-        when specific commands are started and when they finish. The invocation
-        of the :attr:`start_event` and :attr:`finish_event` callbacks is
-        performed inside the :class:`ExternalCommand` class though, so you're
-        free to repurpose these callbacks outside the context of command
-        pools.
-        """
+        """Optional callback that's called just before the command is started (see `event callbacks`_)."""
 
     @property
     def stderr(self):
@@ -1185,6 +1261,47 @@ class ExternalCommand(ControllableProcess):
         """
         return False
 
+    def async_fdel(self):
+        """Delete the value of :attr:`asynchronous`."""
+        delattr(self, 'asynchronous')
+
+    def async_fget(self):
+        """Get the value of :attr:`asynchronous`."""
+        return self.asynchronous
+
+    def async_fset(self, value):
+        """Set the value of :attr:`asynchronous`."""
+        self.asynchronous = value
+
+    def check_retry_allowed(self):
+        """
+        Check if retrying is allowed by invoking the :attr:`retry_event` callback.
+
+        :returns: :data:`True` if :attr:`retry_allowed` is :data:`True` and the
+                  :attr:`retry_event` callback didn't return :data:`False`,
+                  otherwise :data:`False`.
+        """
+        if self.failed and self.retry and not self.retry_allowed:
+            # Log a final warning message when we give up on retrying a failed command.
+            self.logger.warning("Giving up on retrying external command that has failed %s! (%s)",
+                                pluralize(self.retry_count + 1, "time"), self)
+            return False
+        elif self.retry_allowed:
+            # When retrying is enabled and applicable we invoke the
+            # `retry_event' callback to check whether the caller
+            # doesn't veto our decision to retry.
+            if self.invoke_event_callback('retry_event') is False:
+                self.logger.warning("External command failed with return code %i, retry canceled by callback (%s).",
+                                    self.returncode, self)
+                self.retry_allowed = False
+                return False
+            else:
+                self.logger.warning("Will retry external command that failed with return code %i (%s).",
+                                    self.returncode, self)
+                return True
+        else:
+            return False
+
     def format_error_message(self, message, *args, **kw):
         """
         Add the command's captured standard output and/or error to an error message.
@@ -1266,11 +1383,11 @@ class ExternalCommand(ControllableProcess):
 
     def start(self):
         """
-        Start execution of the external command.
+        Start execution of the external command (including :attr:`retry` support).
 
         :raises: - :exc:`ExternalCommandFailed` when
                    :attr:`~ExternalCommand.check` is :data:`True`,
-                   :attr:`async` is :data:`False` and the external command
+                   :attr:`asynchronous` is :data:`False` and the external command
                    exits with a nonzero status code.
 
                  - :exc:`ValueError` when the external command is still
@@ -1280,12 +1397,12 @@ class ExternalCommand(ControllableProcess):
 
         This method instantiates a :class:`subprocess.Popen` object based on
         the defaults defined by :class:`ExternalCommand` and the overrides
-        configured by the caller. What happens then depends on :attr:`async`:
+        configured by the caller. What happens then depends on :attr:`asynchronous`:
 
-        - If :attr:`async` is set :func:`start()` starts the external command
+        - If :attr:`asynchronous` is set :func:`start()` starts the external command
           but doesn't wait for it to end (use :func:`wait()` for that).
 
-        - If :attr:`async` isn't set the ``communicate()`` method on the
+        - If :attr:`asynchronous` isn't set the ``communicate()`` method on the
           :attr:`subprocess` object is called to synchronously execute the
           external command.
         """
@@ -1309,11 +1426,53 @@ class ExternalCommand(ControllableProcess):
         kw['stdout'] = self.stdout_stream.prepare_output(self.stdout_file, self.capture)
         kw['stderr'] = (subprocess.STDOUT if self.merge_streams else
                         self.stderr_stream.prepare_output(self.stderr_file, self.capture_stderr))
-        # Let the operator know what's about to happen.
-        self.logger.debug("Executing external command: %s", quote(kw['args']))
+        if self.retry and not self.asynchronous:
+            # Retry a failing synchronous command.
+            while True:
+                self.start_once(check=False, **kw)
+                # Check if we need to break out of the loop.
+                if self.succeeded or not self.retry_allowed:
+                    # Raise an exception (if applicable).
+                    self.check_errors()
+                    # Stop retrying.
+                    break
+        else:
+            # Retrying of asynchronous commands isn't handled by start().
+            self.start_once(**kw)
+
+    def start_once(self, check=None, **kw):
+        """
+        Start execution of the external command (excluding :attr:`retry` support).
+
+        :param check: Override the value of :attr:`check` for the duration of
+                      this call to :func:`start_once()`. Defaults to
+                      :data:`None`.
+        :param kw: The keyword arguments to the :class:`subprocess.Popen`
+                   initializer (prepared by :func:`start()`).
+
+        The code in this internal method used to be the second half
+        of :func:`start()` but was extracted into a separate method
+        so that it can be called more than once, which made it
+        possible to add :attr:`retry` support.
+        """
+        if self.was_started:
+            if self.check_retry_allowed():
+                # Automatically bump the `retry_count' property.
+                self.retry_count += 1
+            else:
+                # Don't retry command not found errors, stop retrying when
+                # the limit on the number of retries is reached and stop
+                # retrying when the `retry_event' callback returns False.
+                return
         # Lightweight reset of internal state.
         for name in 'error_type', 'pid', 'returncode', 'subprocess':
             delattr(self, name)
+        # Let the operator know what's about to happen.
+        if self.retry and self.retry_limit > 0:
+            self.logger.debug("Executing external command (attempt %i/%i): %s",
+                              self.retry_count + 1, self.retry_limit + 1, self)
+        else:
+            self.logger.debug("Executing external command: %s", self)
         # Invoke the start event callback?
         self.invoke_event_callback('start_event')
         # Remember that we called subprocess.Popen() regardless of whether it
@@ -1331,7 +1490,7 @@ class ExternalCommand(ControllableProcess):
                 self.stdout_stream.finalize(b'')
                 self.stderr_stream.finalize(b'')
                 # Cleanup temporary resources and raise the exception (or not).
-                self.wait()
+                self.wait(check=check)
             else:
                 # Don't swallow exceptions we can't handle.
                 raise
@@ -1341,12 +1500,12 @@ class ExternalCommand(ControllableProcess):
             # subprocess.Popen object without losing track of the process ID.
             self.pid = self.subprocess.pid
             # Synchronously wait for the external command to end?
-            if not self.async:
+            if not self.asynchronous:
                 self.logger.debug("Joining synchronous process using subprocess.Popen.communicate() ..")
                 stdout, stderr = self.subprocess.communicate(self.encoded_input)
                 self.stdout_stream.finalize(stdout)
                 self.stderr_stream.finalize(stderr)
-                self.wait()
+                self.wait(check=check)
         finally:
             # Invoke the finish event callback? (only applies when the command
             # is synchronous or the subprocess module raised an exception)
@@ -1363,10 +1522,10 @@ class ExternalCommand(ControllableProcess):
         :param kw: Any keyword arguments are passed on to
                    :func:`~executor.process.ControllableProcess.wait_for_process()`.
         :raises: :exc:`ExternalCommandFailed` when :attr:`check` is
-                 :data:`True`, :attr:`async` is :data:`True` and the external
+                 :data:`True`, :attr:`asynchronous` is :data:`True` and the external
                  command exits with a nonzero status code.
 
-        The :func:`wait()` function is only useful when :attr:`async` is
+        The :func:`wait()` function is only useful when :attr:`asynchronous` is
         :data:`True`, it performs the following steps:
 
         1. If :attr:`was_started` is :data:`False` the :func:`start()` method
@@ -1387,7 +1546,7 @@ class ExternalCommand(ControllableProcess):
         if not self.was_started:
             self.start()
         if self.is_running:
-            self.wait_for_process()
+            self.wait_for_process(**kw)
         self.cleanup()
         self.check_errors(check=check)
 
@@ -1426,7 +1585,7 @@ class ExternalCommand(ControllableProcess):
         Load output captured from the standard output/error streams.
 
         Reads the contents of the temporary file(s) created by :func:`start()`
-        (when :attr:`async` and :attr:`capture` are both set) into memory so
+        (when :attr:`asynchronous` and :attr:`capture` are both set) into memory so
         that the output doesn't get lost when the temporary file is cleaned up
         by :func:`cleanup()`.
         """
@@ -1442,7 +1601,7 @@ class ExternalCommand(ControllableProcess):
 
         - The temporary file(s) used to to buffer the external command's
           :attr:`input`, :attr:`stdout` and :attr:`stderr` (only when
-          :attr:`async` is :data:`True`).
+          :attr:`asynchronous` is :data:`True`).
 
         - File handles to the previously mentioned temporary files and
           :data:`os.devnull` (used to implement the :attr:`silent` option).
@@ -1458,7 +1617,7 @@ class ExternalCommand(ControllableProcess):
         self.stderr_stream.finalize()
         # Prepare to garbage collect the subprocess.Popen object?
         if self.subprocess is not None:
-            if self.async:
+            if self.asynchronous:
                 self.logger.debug("Joining asynchronous process using subprocess.Popen.wait() ..")
                 # Perform a wait to allow the system to release the resources
                 # associated with the child process; if a wait is not performed,
@@ -1467,6 +1626,8 @@ class ExternalCommand(ControllableProcess):
                 # so we don't lose track of it once we allow the subprocess.Popen
                 # object to be garbage collected.
                 self.returncode = self.subprocess.wait()
+                self.logger.debug("Got return code %i from asynchronous process (%s).",
+                                  self.returncode, self)
                 # Invoke the finish event callback?
                 self.invoke_event_callback('finish_event')
             else:
@@ -1474,6 +1635,8 @@ class ExternalCommand(ControllableProcess):
                 # because computing it again after we destroy our reference
                 # to the subprocess.Popen object will be impossible.
                 self.returncode = self.subprocess.returncode
+                self.logger.debug("Got return code %i from synchronous process (%s).",
+                                  self.returncode, self)
             # Destroy our reference to the subprocess.Popen object
             # to allow it to be garbage collected.
             delattr(self, 'subprocess')
@@ -1514,11 +1677,12 @@ class ExternalCommand(ControllableProcess):
         Invoke one of the event callbacks.
 
         :param name: The name of the callback (a string).
+        :returns: The return value of the callback.
         """
         callback = getattr(self, name)
         if callback is not None:
             logger.debug("Invoking %s callback ..", name)
-            callback(self)
+            return callback(self)
 
     def __enter__(self):
         """
@@ -1530,15 +1694,15 @@ class ExternalCommand(ControllableProcess):
         :keyword:`with` statement, the command is automatically started when
         entering the context and terminated when leaving the context.
 
-        If the proces hasn't already been started yet :attr:`async` is
+        If the proces hasn't already been started yet :attr:`asynchronous` is
         automatically set to :data:`True` (if it's not already :data:`True`),
         otherwise the command will have finished execution by the time the body
         of the :keyword:`with` statement is executed (which isn't really all
         that useful :-).
         """
         if not self.was_started:
-            if not self.async:
-                self.async = True
+            if not self.asynchronous:
+                self.asynchronous = True
             self.start()
         return self
 
@@ -1585,8 +1749,9 @@ class ExternalCommand(ControllableProcess):
         To understand why this is useful, consider the following:
 
         - Iteration over file-like objects in Python uses a hidden read-ahead
-          buffer for efficiency. Refer to the :func:`file.next()` documentation
-          for details.
+          buffer for efficiency. Refer to the `file.next()
+          <https://docs.python.org/2/library/stdtypes.html#file.next>`_
+          documentation for details.
 
         - Pipes are file-like objects so if you iterate over them but the
           command doesn't emit a lot of output, the iteration may not produce
@@ -1595,7 +1760,7 @@ class ExternalCommand(ControllableProcess):
         """
         if not self.was_started:
             if not self.buffered:
-                self.async = True
+                self.asynchronous = True
             self.start()
         for is_enabled, value_property in (('capture', 'stdout'), ('capture_stderr', 'stderr')):
             if getattr(self, is_enabled):
@@ -1611,6 +1776,59 @@ class ExternalCommand(ControllableProcess):
                     # I found no way to work around that.
                     return iter(lambda: stream.readline().decode(self.encoding), u'')
 
+    def __str__(self):
+        """
+        Render a human friendly string representation of the external command.
+
+        This special method calls :func:`quote()` on :attr:`command_line` which
+        is used by :mod:`executor` to enable lazy formatting of log messages
+        containing the quoted command line.
+        """
+        return quote(self.command_line)
+
+
+# We define the 'async' -> 'asynchronous' alias once at module import time. The
+# alternative is to take care of this in ExternalCommand.__init__() but there
+# are two reasons why I prefer module import time instead:
+#
+#  - If we handle this in ExternalCommand.__init__() we'd be allocating a new
+#    property object for every ExternalCommand object, whereas now we allocate
+#    a single property object for all ExternalCommand objects.
+#
+#  - If the alias is created by ExternalCommand.__init__() then Sphinx would
+#    never get to see it (because Sphinx looks at classes, not instances).
+setattr(
+    ExternalCommand,
+    'async',
+    property(
+        ExternalCommand.async_fget,
+        ExternalCommand.async_fset,
+        ExternalCommand.async_fdel,
+        """
+        An alias for the :attr:`asynchronous` property.
+
+        In `release 21.0`_ the :attr:`async` property was renamed to
+        :attr:`asynchronous` because Python 3.7 introduces the :keyword:`async`
+        keyword, invalidating its use as an identifier. This alias ensures
+        backwards compatibility with callers that are still using the old
+        :attr:`async` name. If you're wondering which of the two properties you
+        should use:
+
+        - If Python >= 3.7 compatibility is important to you then your only
+          choice is :attr:`asynchronous`.
+
+        - If you're working in a Python < 3.7 code base you can pick either one
+          of the properties, it really doesn't matter.
+
+        - Of course most Python 2 code bases will eventually need to be
+          upgraded to Python 3 and the future proof choice is
+          :attr:`asynchronous`.
+
+        .. _release 21.0: https://executor.readthedocs.io/en/latest/changelog.html#release-21-0-2018-10-07
+        """,
+    ),
+)
+
 
 class CachedStream(object):
 
@@ -1650,7 +1868,7 @@ class CachedStream(object):
                   :class:`subprocess.Popen` as the ``stdin`` argument.
         """
         if self.command.input is not None:
-            if self.command.async and self.command.input is not True:
+            if self.command.asynchronous and self.command.input is not True:
                 # Store the input provided by the caller in a temporary file
                 # and connect the file to the command's standard input stream.
                 self.prepare_temporary_file()
@@ -1685,7 +1903,7 @@ class CachedStream(object):
             self.redirect(file)
             return self.fd
         elif capture or (self.command.silent and not self.command.really_silent):
-            if self.command.async and self.command.buffered:
+            if self.command.asynchronous and self.command.buffered:
                 # Capture the stream to a temporary file.
                 self.prepare_temporary_file()
                 return self.fd
