@@ -1,7 +1,7 @@
 # Programmer friendly subprocess wrapper.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: April 27, 2018
+# Last Change: October 11, 2018
 # URL: https://executor.readthedocs.io
 
 r"""
@@ -14,8 +14,8 @@ simple wrappers for :class:`.ExternalCommand`, :class:`.RemoteCommand` and
 :class:`.SecureChangeRootCommand`.
 
 This allows you to script interaction with external commands in Python and
-perform that interaction on your local system, on a remote system over SSH_ or
-inside a chroot_ using the exact same Python code. `Dependency injection`_ on
+perform that interaction on your local system, on remote systems over SSH_ or
+inside chroots_ using the exact same Python code. `Dependency injection`_ on
 steroids anyone? :-)
 
 Here's a simple example:
@@ -60,7 +60,7 @@ integration tools developed using Python:
  I wrote on the remote system and I'm done!
 
 .. _SSH: https://en.wikipedia.org/wiki/Secure_Shell
-.. _chroot: http://en.wikipedia.org/wiki/Chroot
+.. _chroots: http://en.wikipedia.org/wiki/Chroot
 .. _Dependency injection: http://en.wikipedia.org/wiki/Dependency_injection
 """
 
@@ -71,10 +71,11 @@ import logging
 import multiprocessing
 import os
 import random
+import shlex
 import socket
 
 # External dependencies.
-from humanfriendly.text import dedent, split
+from humanfriendly.text import dedent, pluralize, split
 from property_manager import (
     PropertyManager,
     lazy_property,
@@ -179,18 +180,56 @@ class AbstractContext(PropertyManager):
         """
         The code name of the system's distribution (a lowercased string like ``precise`` or ``trusty``).
 
-        This is the lowercased output of ``lsb_release --short --codename``.
+        How this property is computed depends on the execution context:
+
+        1. When the file ``/etc/lsb-release`` exists and defines the variable
+           ``DISTRIB_CODENAME`` then this is the preferred source (see
+           :attr:`lsb_release_variables`). This was added in response to
+           `issue #10`_ which reported that the ``lsb_release`` program
+           wasn't available in vanilla Ubuntu 18.04 Docker images.
+
+        2. When the program ``lsb_release`` is installed then the output of the
+           command ``lsb_release --short --codename`` is used.
+
+        The returned string is guaranteed to be lowercased, in order to enable
+        reliable string comparison.
+
+        .. _issue #10: https://github.com/xolox/python-executor/issues/10
         """
-        return self.capture('lsb_release', '--short', '--codename', check=False, silent=True).lower()
+        logger.debug("Trying to discover distribution codename using /etc/lsb-release ..")
+        value = self.lsb_release_variables.get('DISTRIB_CODENAME')
+        if not value:
+            logger.debug("Falling back to 'lsb_release --short --codename' ..")
+            value = self.capture('lsb_release', '--short', '--codename', check=False, silent=True)
+        return value.lower()
 
     @lazy_property
     def distributor_id(self):
         """
         The distributor ID of the system (a lowercased string like ``debian`` or ``ubuntu``).
 
-        This is the lowercased output of ``lsb_release --short --id``.
+        How this property is computed depends on the execution context:
+
+        1. When the file ``/etc/lsb-release`` exists and defines the variable
+           ``DISTRIB_ID`` then this is the preferred source (see
+           :attr:`lsb_release_variables`). This was added in response to
+           `issue #10`_ which reported that the ``lsb_release`` program
+           wasn't available in vanilla Ubuntu 18.04 Docker images.
+
+        2. When the program ``lsb_release`` is installed then the output of the
+           command ``lsb_release --short --id`` is used.
+
+        The returned string is guaranteed to be lowercased, in order to enable
+        reliable string comparison.
+
+        .. _issue #10: https://github.com/xolox/python-executor/issues/10
         """
-        return self.capture('lsb_release', '--short', '--id', check=False, silent=True).lower()
+        logger.debug("Trying to discover distributor ID using /etc/lsb-release ..")
+        value = self.lsb_release_variables.get('DISTRIB_ID')
+        if not value:
+            logger.debug("Falling back to 'lsb_release --short --id' ..")
+            value = self.capture('lsb_release', '--short', '--id', check=False, silent=True)
+        return value.lower()
 
     @lazy_property
     def have_ionice(self):
@@ -203,6 +242,53 @@ class AbstractContext(PropertyManager):
         prototype = self.prepare('true')
         return prototype.have_superuser_privileges or prototype.sudo
 
+    @lazy_property
+    def lsb_release_variables(self):
+        """
+        The contents of ``/etc/lsb-release`` as a dictionary.
+
+        The values of :attr:`distributor_id` and :attr:`distribution_codename`
+        are based on the information provided by :attr:`lsb_release_variables`.
+        If ``/etc/lsb-release`` doesn't exist or can't be parsed a debug
+        message is logged and an empty dictionary is returned. Here's an
+        example:
+
+        >>> from executor.contexts import LocalContext
+        >>> context = LocalContext()
+        >>> context.lsb_release_variables
+        {'DISTRIB_CODENAME': 'bionic',
+         'DISTRIB_DESCRIPTION': 'Ubuntu 18.04.1 LTS',
+         'DISTRIB_ID': 'Ubuntu',
+         'DISTRIB_RELEASE': '18.04'}
+        """
+        variables = dict()
+        # We proceed under the assumption that the file exists, but avoid
+        # raising an exception when it doesn't and we don't leak error messages
+        # to the standard error stream. We could have used is_file() and
+        # is_readable() to "ask for permission instead of forgiveness" (so to
+        # speak) but that requires the execution of three external commands
+        # instead of one to accomplish the exact same thing :-P.
+        logger.debug("Trying to read /etc/lsb-release ..")
+        contents = self.capture('cat', '/etc/lsb-release', check=False, silent=True)
+        logger.debug("Parsing /etc/lsb-release contents: %r", contents)
+        for lnum, line in enumerate(contents.splitlines()):
+            name, delimiter, value = line.partition('=')
+            parsed_value = shlex.split(value)
+            # The zero byte check below guards against a weird edge that has so
+            # far only manifested in the Python 2.6 environment of Travis CI:
+            # The parsing of /etc/lsb-release results in the expected variable
+            # names but values containing binary data including nul bytes.
+            # https://github.com/xolox/python-executor/issues/15.
+            if len(parsed_value) == 1 and '\0' not in parsed_value[0]:
+                variables[name.strip()] = parsed_value[0]
+            else:
+                logger.debug("Failed to parse line %i: %r", lnum + 1, line)
+        if variables:
+            logger.debug("Extracted %s from /etc/lsb-release: %r", pluralize(len(variables), "variable"), variables)
+        else:
+            logger.debug("Failed to read /etc/lsb-release ..")
+        return variables
+
     @writable_property
     def options(self):
         """The options that are passed to commands created by the context (a dictionary)."""
@@ -428,8 +514,8 @@ class AbstractContext(PropertyManager):
 
         - This method *emulates* filename globbing as supported by system
           shells like Bash and ZSH. It works by forking a Python interpreter
-          and using that to call the :mod:`glob.glob()` function. This approach
-          is of course rather heavyweight.
+          and using that to call the :func:`glob.glob()` function. This
+          approach is of course rather heavyweight.
 
         - Initially this method used Bash for filename matching (similar to
           `this StackOverflow answer <https://unix.stackexchange.com/a/34012/44309>`_)
@@ -609,11 +695,12 @@ class AbstractContext(PropertyManager):
         options.update(shell=False, tty=True)
         return self.prepare(DEFAULT_SHELL, **options)
 
-    def read_file(self, filename):
+    def read_file(self, filename, **options):
         """
         Read the contents of a file.
 
         :param filename: The pathname of the file to read (a string).
+        :param options: Optional keyword arguments to :func:`execute()`.
         :returns: The contents of the file (a byte string).
 
         This method uses cat_ to read the contents of files so that options
@@ -623,7 +710,8 @@ class AbstractContext(PropertyManager):
 
         .. _cat: http://linux.die.net/man/1/cat
         """
-        return self.execute('cat', filename, capture=True).stdout
+        options.update(capture=True)
+        return self.execute('cat', filename, **options).stdout
 
     def start_interactive_shell(self, **options):
         """
@@ -662,12 +750,13 @@ class AbstractContext(PropertyManager):
         cmd.start()
         return cmd.succeeded
 
-    def write_file(self, filename, contents):
+    def write_file(self, filename, contents, **options):
         """
         Change the contents of a file.
 
         :param filename: The pathname of the file to write (a string).
         :param contents: The contents to write to the file (a byte string).
+        :param options: Optional keyword arguments to :func:`execute()`.
 
         This method uses a combination of cat_ and `output redirection`_ to
         change the contents of files so that options like
@@ -679,7 +768,8 @@ class AbstractContext(PropertyManager):
 
         .. _output redirection: https://en.wikipedia.org/wiki/Redirection_(computing)
         """
-        return self.execute('cat > %s' % quote(filename), shell=True, input=contents)
+        options.update(input=contents, shell=True)
+        return self.execute('cat > %s' % quote(filename), **options)
 
 
 class LocalContext(AbstractContext):
@@ -712,9 +802,10 @@ class LocalContext(AbstractContext):
         interpreter.
 
         This optimization is skipped when :attr:`~AbstractContext.options`
-        contains :func:`~ExternalCommand.sudo`, :func:`~ExternalCommand.uid` or
-        :func:`~ExternalCommand.user` to avoid reporting wrong matches due to
-        insufficient filesystem permissions.
+        contains :attr:`~executor.ExternalCommand.sudo`,
+        :attr:`~executor.ExternalCommand.uid` or
+        :attr:`~executor.ExternalCommand.user` to avoid reporting wrong matches
+        due to insufficient filesystem permissions.
         """
         if any(map(self.options.get, ('sudo', 'uid', 'user'))):
             return super(LocalContext, self).glob(pattern)
