@@ -1,7 +1,7 @@
 # Automated tests for the `executor' module.
 #
 # Author: Peter Odding <peter@peterodding.com>
-# Last Change: May 3, 2018
+# Last Change: October 7, 2018
 # URL: https://executor.readthedocs.io
 
 """
@@ -55,9 +55,11 @@ import time
 import uuid
 
 # External dependencies.
-from humanfriendly import Timer, compact
+from humanfriendly import Timer, coerce_boolean, compact, dedent
 from humanfriendly.testing import TemporaryDirectory, TestCase, retry, run_cli
 from mock import MagicMock
+from property_manager import set_property
+from six.moves import StringIO
 
 # Modules included in our package.
 from executor import (
@@ -115,6 +117,19 @@ class ExecutorTestCase(TestCase):
         if not os.path.isdir(self.sudo_enabled_directory):
             os.makedirs(self.sudo_enabled_directory)
 
+    def test_async_compat(self):
+        """Make sure the ``async`` property still works (backwards compatibility)."""
+        aliases = ('async', 'asynchronous')
+        cmd = ExternalCommand('test', **{'async': True})
+        assert all(getattr(cmd, n) is True for n in aliases)
+        cmd = ExternalCommand('test', **{'async': False})
+        assert all(getattr(cmd, n) is False for n in aliases)
+
+    def test_double_start(self):
+        """Make sure a command can't be started when it's already running."""
+        with ExternalCommand('sleep', '1') as cmd:
+            self.assertRaises(ValueError, cmd.start)
+
     def test_graceful_termination(self):
         """Test graceful termination of processes."""
         self.check_termination(method='terminate')
@@ -198,10 +213,10 @@ class ExecutorTestCase(TestCase):
         self.assertEqual(e.returncode, 42)
         self.assertTrue('whatever' in e.error_message)
         # Make sure the CommandNotFound exception is raised consistently
-        # regardless of the values of the `shell' and `async' options.
-        for async in True, False:
+        # regardless of the values of the `shell' and `asynchronous' options.
+        for asynchronous in True, False:
             for shell in True, False:
-                cmd = ExternalCommand(MISSING_COMMAND, async=async, shell=shell)
+                cmd = ExternalCommand(MISSING_COMMAND, asynchronous=asynchronous, shell=shell)
                 self.assertRaises(CommandNotFound, cmd.wait)
                 assert cmd.returncode == COMMAND_NOT_FOUND_STATUS
 
@@ -313,6 +328,18 @@ class ExecutorTestCase(TestCase):
             lines = [line.strip() for line in handle]
         assert lines == ['existing contents', 'appended output']
 
+    def test_redirect_without_fd(self):
+        """Test redirection to a file object that doesn't have an associated file descriptor."""
+        mock_file = StringIO()
+        setattr(mock_file, 'name', '/some/random/path')
+        self.assertRaises(ValueError, execute, 'true', stdout_file=mock_file)
+
+    def test_redirect_without_name(self):
+        """Test redirection to a file object that doesn't have an associated filename."""
+        mock_file = StringIO()
+        setattr(mock_file, 'fileno', lambda: 5)
+        self.assertRaises(ValueError, execute, 'true', stdout_file=mock_file)
+
     def test_merged_streams_to_file(self):
         """Make sure the standard streams of external commands can be merged, redirected and appended to a file."""
         fd, filename = tempfile.mkstemp(prefix='executor-', suffix='-merged.txt')
@@ -332,7 +359,7 @@ class ExecutorTestCase(TestCase):
         fd, filename = tempfile.mkstemp(prefix='executor-', suffix='-streaming.txt')
         with open(filename, 'w') as handle:
             cmd = ExternalCommand('for ((i=0; i<25; i++)); do echo $i; sleep 0.1; done',
-                                  async=True, stdout_file=handle)
+                                  asynchronous=True, stdout_file=handle)
             cmd.start()
 
         def expect_some_output():
@@ -358,7 +385,7 @@ class ExecutorTestCase(TestCase):
                          'sys.stdout.write(sys.stdin.readline().upper())',
                          'sys.stdout.flush()',
                          'sys.stdout.write(sys.stdin.readline().upper())'),
-            async=True, buffered=False, capture=True, input=True
+            asynchronous=True, buffered=False, capture=True, input=True
         )
         with cmd:
             # Message the command.
@@ -468,7 +495,7 @@ class ExecutorTestCase(TestCase):
     def test_ionice_option(self):
         """Make sure ``ionice`` can be used."""
         rsync_command_line = ['rsync', '-a', '/', '/mnt/backups/latest/']
-        expected_ionice_command = ['ionice', '--class', 'idle']
+        expected_ionice_command = ['ionice', '-c', 'idle']
         command = ExternalCommand(*rsync_command_line, ionice='idle')
         assert command.ionice == 'idle'
         print(command.ionice_command)
@@ -503,9 +530,9 @@ class ExecutorTestCase(TestCase):
         cmd.wait()
         assert cmd.output == 'Also works fine'
 
-    def test_simple_async_cmd(self):
+    def test_simple_asynchronous_cmd(self):
         """Make sure commands can be executed asynchronously."""
-        cmd = ExternalCommand('sleep 4', async=True)
+        cmd = ExternalCommand('sleep 4', asynchronous=True)
         # Make sure we're starting from a sane state.
         assert not cmd.was_started
         assert not cmd.is_running
@@ -534,11 +561,11 @@ class ExecutorTestCase(TestCase):
         assert cmd.is_finished
         assert cmd.returncode == 0
 
-    def test_async_with_input(self):
+    def test_asynchronous_with_input(self):
         """Make sure asynchronous commands can be provided standard input."""
-        random_file = os.path.join(tempfile.gettempdir(), 'executor-%s-async-input-test' % os.getpid())
+        random_file = os.path.join(tempfile.gettempdir(), 'executor-%s-asynchronous-input-test' % os.getpid())
         random_value = str(random.random())
-        cmd = ExternalCommand('cat > %s' % quote(random_file), async=True, input=random_value)
+        cmd = ExternalCommand('cat > %s' % quote(random_file), asynchronous=True, input=random_value)
         try:
             cmd.start()
             cmd.wait()
@@ -550,10 +577,10 @@ class ExecutorTestCase(TestCase):
             if os.path.isfile(random_file):
                 os.unlink(random_file)
 
-    def test_async_with_output(self):
+    def test_asynchronous_with_output(self):
         """Make sure asynchronous command output can be captured."""
         random_value = str(random.random())
-        cmd = ExternalCommand('echo %s' % quote(random_value), async=True, capture=True)
+        cmd = ExternalCommand('echo %s' % quote(random_value), asynchronous=True, capture=True)
         cmd.start()
         cmd.wait()
         assert cmd.output == random_value
@@ -569,11 +596,11 @@ class ExecutorTestCase(TestCase):
 
     def test_event_callbacks(self):
         """Make sure the ``start_event`` and ``finish_event`` callbacks are actually invoked."""
-        for async in True, False:
+        for asynchronous in True, False:
             results = []
             cmd = ExternalCommand(
                 'sleep', '0.1',
-                async=async,
+                asynchronous=asynchronous,
                 start_event=lambda cmd: results.append(('started', time.time())),
                 finish_event=lambda cmd: results.append(('finished', time.time())),
             )
@@ -588,14 +615,14 @@ class ExecutorTestCase(TestCase):
     def test_repr(self):
         """Make sure that repr() on external commands gives sane output."""
         cmd = ExternalCommand('echo 42',
-                              async=True,
+                              asynchronous=True,
                               capture=True,
                               directory='/',
                               environment={'my_environment_variable': '42'})
         assert repr(cmd).startswith('ExternalCommand(')
         assert repr(cmd).endswith(')')
         assert 'echo 42' in repr(cmd)
-        assert 'async=True' in repr(cmd)
+        assert 'asynchronous=True' in repr(cmd)
         assert ('directory=%r' % '/') in repr(cmd)
         assert 'my_environment_variable' in repr(cmd)
         assert 'was_started=False' in repr(cmd)
@@ -611,6 +638,44 @@ class ExecutorTestCase(TestCase):
 
         retry(assert_finished, 10)
 
+    def test_retry(self):
+        """Check that failing commands can be retried until they succeed."""
+        with TemporaryDirectory() as directory:
+            script = self.create_retry_script(directory, 5)
+            cmd = ExternalCommand(script, retry=True, retry_limit=10, shell=False)
+            cmd.start()
+            assert cmd.retry_count == 4
+            assert cmd.returncode == 0
+
+    def test_retry_limit(self):
+        """Check that failing commands aren't retried indefinitely."""
+        with TemporaryDirectory() as directory:
+            script = self.create_retry_script(directory, 5)
+            cmd = ExternalCommand(script, check=False, retry=True, retry_limit=2, shell=False)
+            cmd.start()
+            assert cmd.retry_count == 2
+            assert cmd.returncode == 42
+
+    def create_retry_script(self, directory, iterations=2):
+        """Create a script that fails until the fifth run :-)."""
+        unique_name = uuid.uuid4().hex
+        script = os.path.join(directory, '%s.sh' % unique_name)
+        data_file = os.path.join(directory, '%s.txt' % unique_name)
+        with open(script, 'w') as handle:
+            handle.write(dedent('''
+                #!/bin/bash -e
+
+                ITERATION=$(cat {data_file} 2>/dev/null || echo 1)
+                echo $(($ITERATION + 1)) > {data_file}
+                if [ $ITERATION -ge {limit} ]; then
+                    exit 0
+                else
+                    exit 42
+                fi
+            ''', data_file=data_file, limit=iterations))
+        os.chmod(script, 0o777)
+        return script
+
     def test_command_pool(self):
         """Make sure command pools actually run multiple commands in parallel."""
         num_commands = 10
@@ -643,6 +708,45 @@ class ExecutorTestCase(TestCase):
         e2 = intercept(ExternalCommandFailed, pool.collect)
         assert e2.command is c2
 
+    def test_command_pool_retry(self):
+        """Make sure command pools can retry failing commands."""
+        with TemporaryDirectory() as directory:
+            pool = CommandPool(concurrency=2, delay_checks=True)
+            # Create a shell script that succeeds on the second run and retry
+            # it exactly once. We expect this command to have succeeded when
+            # the command pool is finished.
+            script_1 = self.create_retry_script(directory, 2)
+            command_1 = ExternalCommand(script_1, asynchronous=True, retry=True, retry_limit=1)
+            pool.add(command_1)
+            # Create a shell script that succeeds on the fourth run and retry
+            # it up to two times. We expect this command to have failed when
+            # the command pool is finished.
+            script_2 = self.create_retry_script(directory, 4)
+            command_2 = ExternalCommand(script_2, asynchronous=True, retry=True, retry_limit=2)
+            pool.add(command_2)
+            # Include a command without retries that succeeds.
+            command_3 = ExternalCommand('true', asynchronous=True, retry=False)
+            pool.add(command_3)
+            # Include a command without retries that fails.
+            command_4 = ExternalCommand('false', asynchronous=True, retry=False)
+            pool.add(command_4)
+            # Run the commands in the pool, expecting an `CommandPoolFailed'
+            # exception because the second command will fail despite retrying
+            # and the fourth command fails on its first and only run.
+            self.assertRaises(CommandPoolFailed, pool.run)
+            # Check that the first command succeeded (with a retry).
+            assert command_1.succeeded
+            assert command_1.retry_count == 1
+            # Check that the second command failed (with retries).
+            assert command_2.failed
+            assert command_2.retry_count == 2
+            # Check that the third command succeeded (without retries).
+            assert command_3.succeeded
+            assert command_3.retry_count == 0
+            # Check that the fourth command failed (without retries).
+            assert command_4.failed
+            assert command_4.retry_count == 0
+
     def test_command_pool_termination(self):
         """Make sure command pools can be terminated on failure."""
         pool = CommandPool()
@@ -915,15 +1019,33 @@ class ExecutorTestCase(TestCase):
 
     def test_lsb_release_shortcuts(self):
         """Test the ``lsb_release`` shortcuts."""
+        # The following tests should pass on my laptops and Travis CI, but I
+        # don't want this test to fail on `unexpected' platforms, so here's a
+        # pragmatic compromise :-).
         try:
-            # The following tests should pass on my laptops and Travis CI.
             context = LocalContext()
             assert context.distributor_id == 'ubuntu'
-            assert context.distribution_codename in ('precise', 'trusty', 'xenial')
-        except AssertionError:
-            # But I don't want this test to fail on `unexpected'
-            # platforms so here's a pragmatic compromise :-).
-            return self.skipTest("assuming unsupported platform")
+            assert context.distribution_codename in ('precise', 'trusty', 'xenial', 'bionic')
+        except Exception:
+            if self.lsb_release_expected_to_work:
+                raise
+            else:
+                return self.skipTest("assuming unsupported platform")
+
+    def test_lsb_release_fallback(self):
+        """Make sure the fall back from /etc/lsb-release to /usr/bin/lsb_release works."""
+        context = LocalContext()
+        # Disable the parsing of /etc/lsb-release.
+        set_property(context, 'lsb_release_variables', {})
+        # Test the fall back from /etc/lsb-release to /usr/bin/lsb_release.
+        try:
+            assert context.distributor_id == 'ubuntu'
+            assert context.distribution_codename in ('precise', 'trusty', 'xenial', 'bionic')
+        except Exception:
+            if self.lsb_release_expected_to_work:
+                raise
+            else:
+                return self.skipTest("assuming unsupported platform")
 
     def test_lsb_release_error_handling(self):
         """Test that the ``lsb_release`` shortcuts don't raise exceptions on unsupported platforms."""
@@ -931,6 +1053,20 @@ class ExecutorTestCase(TestCase):
         assert context.distributor_id == ''
         assert context.distribution_codename == ''
 
+    def test_lsb_release_variables(self):
+        """Test the ``/etc/lsb-release`` parsing."""
+        try:
+            context = LocalContext()
+            variables = context.lsb_release_variables
+            assert len(variables) > 0
+            assert variables['DISTRIB_ID'].lower() == 'ubuntu'
+            assert variables['DISTRIB_CODENAME'].lower() in ('precise', 'trusty', 'xenial', 'bionic')
+        except Exception:
+            if self.lsb_release_expected_to_work:
+                raise
+            else:
+                return self.skipTest("assuming unsupported platform")
+
     def test_local_context(self):
         """Test a local command context."""
         self.check_context(LocalContext())
@@ -1101,6 +1237,17 @@ class ExecutorTestCase(TestCase):
             assert timer.elapsed_time < 10
         retry(timeout_hammer, 60)
 
+    @property
+    def lsb_release_expected_to_work(self):
+        """True on my laptops and Travis CI, False otherwise."""
+        developer_laptop = (socket.gethostname() in ('peter-mba', 'peter-mbp'))
+        travis_ci = coerce_boolean(os.environ.get('TRAVIS', 'false'))
+        # TODO For some weird reason the parsing of the /etc/lsb-release file
+        # in the Travis CI build environment for Python 2.6 seems to result in
+        # binary garbage. Right now I'm not inclined to dive deeper.
+        py26 = (sys.version_info[:2] == (2, 6))
+        return developer_laptop or (travis_ci and not py26)
+
 
 def intercept(exc_type, func, *args, **kw):
     """Intercept and return a raised exception."""
